struct SolverResults

    solution_method::Symbol
    initial::Array{Float64,1}
    zero::Array{Float64,1}
    fzero::Array{Float64,1}
    xdist::Float64
    fdist::Float64
    iters::Integer

end

function box_projection(x::Array{T,1},l::Array{T,1},u::Array{T,1}) where {T <: AbstractFloat}

    y = copy(x)

    for i in eachindex(x)
        if y[i] < l[i]
            y[i] = l[i]
        elseif y[i] > u[i]
            y[i] = u[i]
        end
    end

    return y

end

function box_projection!(x::Array{T,1},l::Array{T,1},u::Array{T,1}) where {T <: AbstractFloat}

    for i in eachindex(x)
        if x[i] < l[i]
            x[i] = l[i]
        elseif x[i] > u[i]
            x[i] = u[i]
        end
    end

end

function constrained_newton(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100) where {T <: AbstractFloat, S <: Integer}

    n = length(x)
    xk = copy(x)
    xn = similar(x)
    jk = Array{T}(undef,n,n)

    lenx = zero(T)
    lenf = zero(T)

    iter = 0
    while true

        jk .= ForwardDiff.jacobian(f,xk)
        xn .= xk - jk\f(xk)
  
        box_projection!(xn,l,u)

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))
    
        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end
  
    end
  
    results = SolverResults(:newton,x,xn,f(xn),lenx,lenf,iter)

    return results
  
end

function constrained_newton_sparse(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100) where {T <: AbstractFloat, S <: Integer}

    n = length(x)
    xk = copy(x)
    xn = similar(x)
    jk = sparse(Array{T}(undef,n,n))

    lenx = zero(T)
    lenf = zero(T)

    iter = 0
    while true

        jk .= sparse(ForwardDiff.jacobian(f,xk))
        xn .= xk - jk\f(xk)
  
        box_projection!(xn,l,u)

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))
    
        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end
  
    end
  
    results = SolverResults(:newton,x,xn,f(xn),lenx,lenf,iter)

    return results
  
end

function constrained_levenberg_marquardt(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100) where {T <: AbstractFloat, S <: Integer}

    n = length(x)
    xk = copy(x)
    xn = similar(x)
    dk = similar(x)
    jk = Array{T}(undef,n,n)

    lenx = zero(T)
    lenf = zero(T)

    muk = (1/2)*10^(-8)*norm(f(xk))^2

    iter = 0
    while true

        jk .= ForwardDiff.jacobian(f,xk)
        dk .= -(jk'jk + muk*I)\(jk'f(xk))
        xn .= xk + dk

        box_projection!(xn,l,u)

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))
  
        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

        muk = min(muk,norm(f(xk))^2)

    end
  
    results = SolverResults(:lm,x,xn,f(xn),lenx,lenf,iter)

    return results
   
end

function constrained_levenberg_marquardt_sparse(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100) where {T <: AbstractFloat, S <: Integer}

    n = length(x)
    xk = copy(x)
    xn = similar(x)
    dk = similar(x)
    jk = sparse(Array{T}(undef,n,n))

    lenx = zero(T)
    lenf = zero(T)

    muk = (1/2)*10^(-8)*norm(f(xk))^2

    iter = 0
    while true

        jk .= sparse(ForwardDiff.jacobian(f,xk))
        dk .= -(jk'jk + muk*I)\(jk'f(xk))
        xn .= xk + dk

        box_projection!(xn,l,u)

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))
  
        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

        muk = min(muk,norm(f(xk))^2)

    end
  
    results = SolverResults(:lm,x,xn,f(xn),lenx,lenf,iter)

    return results
   
end

function constrained_levenberg_marquardt_kyf(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100) where {T <: AbstractFloat, S <: Integer}

    # This is an implementation of Algorithm 3.12 from Kanzow, Yamashita, and Fukushima (2004) "Levenberg-Marquardt methods
    # with strong local convergence properties for solving nonlinear equations with convex constraints", Journal of 
    # Computational and Applied Mathematics, 172, pp. 375--397.

    n = length(x)
    xk = copy(x)
    xn = similar(x)
    xt = similar(x)
    dk = similar(x)
    g =  similar(x)
    jk = Array{T}(undef,n,n)

    lenx = zero(T)
    lenf = zero(T)

    beta  = 0.9
    gamma = 0.99995
    sigma = 10^(-4)
    rho   = 10^(-8)
    p     = 2.1

    muk = (1/2)*10^(-8)*norm(f(xk))^2

    iter = 0
    while true

        jk .= ForwardDiff.jacobian(f,xk)
        dk .= -(jk'jk + muk*I)\(jk'f(xk))
        xn .= xk + dk
        box_projection!(xn,l,u)

        if norm(f(xn)) > gamma*norm(f(xk))
            g .= jk'f(xk)
            if g'dk <= -rho*norm(dk)^p
                alpha = 1.0
                while norm(f(xk+alpha*dk))^2 > norm(f(xk))^2 + 2*alpha*beta*g'dk
                    alpha = beta*alpha
                end
                xn .= xk + alpha*dk
            else
                alpha = 1.0
                while true
                    xt .= xk-alpha*g
                    box_projection!(xt,l,u)
                    if norm(f(xt))^2 <= norm(f(xk))^2 + 2*sigma*g'(xt-xk)
                        xn .=  xt
                        break
                    else
                        alpha = beta*alpha
                    end
                end
            end
        end

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))
  
        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

        muk = min(muk,norm(f(xk))^2)

    end
  
    results = SolverResults(:lm_kyf,x,xn,f(xn),lenx,lenf,iter)

    return results
   
end

function constrained_levenberg_marquardt_kyf_sparse(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100) where {T <: AbstractFloat, S <: Integer}

    # This is an implementation of Algorithm 3.12 from Kanzow, Yamashita, and Fukushima (2004) "Levenberg-Marquardt methods
    # with strong local convergence properties for solving nonlinear equations with convex constraints", Journal of 
    # Computational and Applied Mathematics, 172, pp. 375--397.

    n = length(x)
    xk = copy(x)
    xn = similar(x)
    xt = similar(x)
    dk = similar(x)
    g =  similar(x)
    jk = sparse(Array{T}(undef,n,n))

    lenx = zero(T)
    lenf = zero(T)

    beta  = 0.9
    gamma = 0.99995
    sigma = 10^(-4)
    rho   = 10^(-8)
    p     = 2.1

    muk = (1/2)*10^(-8)*norm(f(xk))^2

    iter = 0
    while true

        jk .= sparse(ForwardDiff.jacobian(f,xk))
        dk .= -(jk'jk + muk*I)\(jk'f(xk))
        xn .= xk + dk
        box_projection!(xn,l,u)

        if norm(f(xn)) > gamma*norm(f(xk))
            g .= jk'f(xk)
            if g'dk <= -rho*norm(dk)^p
                alpha = 1.0
                while norm(f(xk+alpha*dk))^2 > norm(f(xk))^2 + 2*alpha*beta*g'dk
                    alpha = beta*alpha
                end
                xn .= xk + alpha*dk
            else
                alpha = 1.0
                while true
                    xt .= xk-alpha*g
                    box_projection!(xt,l,u)
                    if norm(f(xt))^2 <= norm(f(xk))^2 + 2*sigma*g'(xt-xk)
                        xn .=  xt
                        break
                    else
                        alpha = beta*alpha
                    end
                end
            end
        end

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))
  
        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

        muk = min(muk,norm(f(xk))^2)

    end
  
    results = SolverResults(:lm_kyf,x,xn,f(xn),lenx,lenf,iter)

    return results
   
end

function constrained_levenberg_marquardt_fan(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100) where {T <: AbstractFloat, S <: Integer}

    # This is an implementation of Algorithm 2.1 from Fan (2013) "On the Levenberg-Marquardt methods for convex constrained
    # nonlinear equations", Journal of Industrial and Management Optimization, 9, 1, pp. 227--241.

    n = length(x)
    xk = copy(x)
    xn = similar(x)
    xt = similar(x)
    dk = similar(x)
    jk = Array{T}(undef,n,n)

    lenx = zero(T)
    lenf = zero(T)

    delta = 1.0
    mu    = 10^(-4)
    gamma = 0.99995
    beta  = 0.9
    sigma = 10^(-4)

    iter = 0
    while true

        muk = mu*norm(f(xk))^delta

        jk .= ForwardDiff.jacobian(f,xk)
        dk .= -(jk'jk + muk*I)\(jk'f(xk))
        xn .= xk+dk
        box_projection!(xn,l,u)

        if norm(f(xn)) > gamma*norm(f(xk))
            alpha = 1.0
            while true
                xt .= xk-alpha*jk'f(xk)
                box_projection!(xt,l,u)
                if norm(f(xt))^2 <= norm(f(xk))^2 + sigma*(f(xk)'jk*(xt-xk))
                    xn .= xt
                    break
                else
                    alpha = beta*alpha
                end
            end
        end

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))
  
        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

    end
  
    results = SolverResults(:lm_fan,x,xn,f(xn),lenx,lenf,iter)

    return results
 
end

function constrained_levenberg_marquardt_fan_sparse(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100) where {T <: AbstractFloat, S <: Integer}

    # This is an implementation of Algorithm 2.1 from Fan (2013) "On the Levenberg-Marquardt methods for convex constrained
    # nonlinear equations", Journal of Industrial and Management Optimization, 9, 1, pp. 227--241.

    n = length(x)
    xk = copy(x)
    xn = similar(x)
    xt = similar(x)
    dk = similar(x)
    jk = sparse(Array{T}(undef,n,n))

    lenx = zero(T)
    lenf = zero(T)

    delta = 1.0
    mu    = 10^(-4)
    gamma = 0.99995
    beta  = 0.9
    sigma = 10^(-4)

    iter = 0
    while true

        muk = mu*norm(f(xk))^delta

        jk .= sparse(ForwardDiff.jacobian(f,xk))
        dk .= -(jk'jk + muk*I)\(jk'f(xk))
        xn .= xk+dk
        box_projection!(xn,l,u)

        if norm(f(xn)) > gamma*norm(f(xk))
            alpha = 1.0
            while true
                xt .= xk-alpha*jk'f(xk)
                box_projection!(xt,l,u)
                if norm(f(xt))^2 <= norm(f(xk))^2 + sigma*(f(xk)'jk*(xt-xk))
                    xn .= xt
                    break
                else
                    alpha = beta*alpha
                end
            end
        end

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))
  
        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

    end
  
    results = SolverResults(:lm_fan,x,xn,f(xn),lenx,lenf,iter)

    return results
 
end

function constrained_levenberg_marquardt_ar(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100) where {T <: AbstractFloat, S <: Integer}

    # This is an implementation of Algorithm 2.1 from Amini and Rostami (2016), "Three-steps modified Levenberg-Marquardt 
    # method with a new line search for systems of nonlinear equations", Journal of Computational and Applied Mathematics, 
    # 300, pp. 30--42.

    # Modified to allow for box-constraints by Richard Dennis.

    n = length(x)
    xk = copy(x)
    xn = similar(x)
    z  = similar(x)
    s  = similar(x)
    jk = Array{T}(undef,n,n)

    lenx = zero(T)
    lenf = zero(T)

    sigma1 = 0.005
    sigma2 = 0.005
    rho    = 0.8
    r      = 0.5
    gamma  = 10^(-16)
    mu     = 10^(-4)

    iter = 0
    while true

        jk .= ForwardDiff.jacobian(f,xk)
        lambdak = mu*norm(f(xk))^2

        d1k = -(jk'jk + lambdak*I)\(jk'f(xk))
        d2k = -(jk'jk + lambdak*I)\(jk'f(xk+d1k))
        d3k = -(jk'jk + lambdak*I)\(jk'f(xk+d1k+d2k))

        dk = d1k+d2k+d3k

        z .= xk+dk
        box_projection!(z,l,u)
        s .= z-xk

        if norm(f(z)) <= rho*norm(f(xk))
            alpha = 1.0
        else
            if f(xk)'jk*dk > -gamma
                dk = d1k
                z .= xk+dk
                box_projection!(z,l,u)
                s .= z-xk
            end
            alpha = 1.0
            epsilon = 1/10
            while true
                if norm(f(xk+alpha*s))^2 > (1+epsilon)*norm(f(xk))^2 - sigma1*alpha^2*norm(s)^2 - sigma2*alpha^2*norm(f(xk))^2
                    alpha = r*alpha
                    epsilon = r*epsilon
                else
                    break
                end
            end
        end

        xn .= xk + alpha*s

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))
  
        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

    end
  
    results = SolverResults(:lm_ar,x,xn,f(xn),lenx,lenf,iter)

    return results
  
end

function constrained_levenberg_marquardt_ar_sparse(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100) where {T <: AbstractFloat, S <: Integer}

    # This is an implementation of Algorithm 2.1 from Amini and Rostami (2016), "Three-steps modified Levenberg-Marquardt 
    # method with a new line search for systems of nonlinear equations", Journal of Computational and Applied Mathematics, 
    # 300, pp. 30--42.

    # Modified to allow for box-constraints by Richard Dennis.

    n = length(x)
    xk = copy(x)
    xn = similar(x)
    z  = similar(x)
    s  = similar(x)
    jk = sparse(Array{T}(undef,n,n))

    lenx = zero(T)
    lenf = zero(T)

    sigma1 = 0.005
    sigma2 = 0.005
    rho    = 0.8
    r      = 0.5
    gamma  = 10^(-16)
    mu     = 10^(-4)

    iter = 0
    while true

        jk .= sparse(ForwardDiff.jacobian(f,xk))
        lambdak = mu*norm(f(xk))^2

        d1k = -(jk'jk + lambdak*I)\(jk'f(xk))
        d2k = -(jk'jk + lambdak*I)\(jk'f(xk+d1k))
        d3k = -(jk'jk + lambdak*I)\(jk'f(xk+d1k+d2k))

        dk = d1k+d2k+d3k

        z .= xk+dk
        box_projection!(z,l,u)
        s .= z-xk

        if norm(f(z)) <= rho*norm(f(xk))
            alpha = 1.0
        else
            if f(xk)'jk*dk > -gamma
                dk = d1k
                z .= xk+dk
                box_projection!(z,l,u)
                s .= z-xk
            end
            alpha = 1.0
            epsilon = 1/10
            while true
                if norm(f(xk+alpha*s))^2 > (1+epsilon)*norm(f(xk))^2 - sigma1*alpha^2*norm(s)^2 - sigma2*alpha^2*norm(f(xk))^2
                    alpha = r*alpha
                    epsilon = r*epsilon
                else
                    break
                end
            end
        end

        xn .= xk + alpha*s

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))
  
        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

    end
  
    results = SolverResults(:lm_ar,x,xn,f(xn),lenx,lenf,iter)

    return results
  
end
#=
function coleman_li(f::Function,j::Array{T,2},x::Array{T,1},l::Array{T,1},u::Array{T,1}) where {T <: AbstractFloat}

    n = length(x)
    D = zeros(n,n)

    df = j'*f(x)

    for i = 1:n
        if df[i] < 0.0 && u[i] < Inf
            D[i,i] = u[i] - x[i]
        elseif df[i] > 0.0 && l[i] > -Inf
            D[i,i] = x[i] - l[i]
        elseif df[i] == 0.0 && (l[i] > -Inf || u[i] < Inf)
            D[i,i] = min(x[i]-l[i],u[i]-x[i])
        else
            D[i,i] = 1.0
        end
    end

    return D

end
=#
function coleman_li(f::Function,j::AbstractArray{T,2},x::Array{T,1},l::Array{T,1},u::Array{T,1}) where {T <: AbstractFloat}

    df = j'f(x)

    for i in eachindex(df)
        if df[i] < 0.0 && u[i] < Inf
            df[i] = u[i] - x[i]
        elseif df[i] > 0.0 && l[i] > -Inf
            df[i] = x[i] - l[i]
        elseif df[i] == 0.0 && (l[i] > -Inf || u[i] < Inf)
            df[i] = min(x[i]-l[i],u[i]-x[i])
        else
            df[i] = 1.0
        end
    end
    return df
end

function step_selection(f::Function,x::Array{T,1},Gk::AbstractArray{T,2},jk::AbstractArray{T,2},gk::Array{T,1},pkn::Array{T,1},l::Array{T,1},u::Array{T,1},deltak::T,theta::T) where {T <: AbstractFloat}

    lambdak = Inf
    for i in eachindex(gk)
        if gk[i] != 0.0
            lambdak = min(max(((l[i] - x[i])/gk[i]),((u[i]-x[i])/gk[i])),lambdak)
        else
            lambdak = minimum((Inf,lambdak))
        end
    end

    taukprime = min(-(f(x)'*jk*gk)/norm(jk*gk)^2,deltak/norm(Gk*gk))
    
    tauk = taukprime
    if !isequal(x+taukprime*gk, box_projection(x+taukprime*gk,l,u))
        tauk = theta*lambdak
    end

    pc = tauk*gk

    pc_pkn = pc-pkn
    
    if norm(-pc_pkn) < 1e-13 # "division by zero errors can otherwise occur"
        return (pc+pkn)/2
    else
        gammahat = -(f(x)+jk*pc)'*jk*(-pc_pkn)/norm(jk*(-pc_pkn))^2
        if (pc'Gk^2*(pc_pkn))^2 - norm(Gk*(pc_pkn))^2*(norm(Gk*pc)^2-deltak^2) >= 0.0
            r = ((pc'Gk^2*(pc_pkn))^2 - norm(Gk*(pc_pkn))^2*(norm(Gk*pc)^2-deltak^2))^(1/2)
            gammaplus  = (pc'*Gk^2*(pc_pkn) + r)/norm(Gk*(pc_pkn))^2
            gammaminus = (pc'*Gk^2*(pc_pkn) - r)/norm(Gk*(pc_pkn))^2
        else
            gammaplus  = pc'*Gk^2*(pc_pkn)/norm(Gk*(pc_pkn))^2
            gammaminus = pc'*Gk^2*(pc_pkn)/norm(Gk*(pc_pkn))^2
        end

        if gammahat > 1.0
            gammatildaplus = Inf
            for i in eachindex(pkn)
                if (pkn[i]-pc[i]) != 0.0
                    gammatildaplus = min(max(((l[i] - x[i]- pc[i])/(pkn[i]-pc[i])),((u[i] - x[i] - pc[i])/(pkn[i]-pc[i]))),gammatildaplus)
                else
                    gammatildaplus = min(Inf,gammatildaplus)
                end
            end
            gamma = min(gammahat,gammaplus,theta*gammatildaplus)
        elseif gammahat < 0.0
            gammatildaminus = -Inf
            for i in eachindex(pkn)
                if (-pkn[i]+pc[i]) != 0.0
                    gammatildaminus = min(max(-((l[i] - x[i]- pc[i])/(-(pkn[i]-pc[i]))),-((u[i] - x[i] - pc[i])/(-(pkn[i]-pc[i])))),gammatildaminus)
                else
                    gammatildaminus = min(Inf,gammatildaminus)
                end
            end
            gamma = max(gammahat,gammaminus,theta*gammatildaminus)
        else
            gamma = gammahat
        end

        p = (1.0-gamma)*pc + gamma*pkn

        return p

    end

end

function constrained_dogleg_solver(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100) where {T <: AbstractFloat, S <: Integer}

    # This is an implementation of the algorithm from Bellavia, Macconi, and Pieraccini (2012), "Constrained 
    # dogleg methods for nonlinear systems with simple bounds", Computational Optimization and Applications, 
    # 53, pp. 771--794 

    n = length(x)
    xk  = copy(x)
    xn  = similar(x)
    pkn = similar(x)
    p   = similar(x)
    df  = similar(x)

    jk = Array{T}(undef,n,n)
    Gk = Array{T}(undef,n,n)
    gk = similar(x)

    lenx = zero(T)
    lenf = zero(T)

    # Replace infinities with largest possible Float64
    
    for i in eachindex(x)
        if l[i] == -Inf
            l[i] = -1/eps(T)
        elseif l[i] == Inf
            l[i] = 1/eps(T)
        end
        if u[i] == -Inf
            u[i] = -1/eps(T)
        elseif u[i] == Inf
            u[i] = 1/eps(T)
        end
    end
    
    deltak = 1.0 
    theta  = 0.99995 
    beta1  = 0.5#0.25 
    beta2  = 0.9 

    iter = 0
    while true

        jk .= ForwardDiff.jacobian(f,xk)
        df .= coleman_li(f,jk,xk,l,u)
        Gk .= Diagonal(df.^(-1/2))
        gk .= -df.*jk'f(xk)
   
        alphak = max(theta,1.0-norm(f(xk)))
        pkn .= xk-jk\f(xk)
        box_projection!(pkn,l,u)
        pkn .= alphak*(pkn-xk)

        p .= step_selection(f,xk,Gk,jk,gk,pkn,l,u,deltak,theta)
    
        while true
            rhof = (norm(f(xk)) - norm(f(xk+p))) / (norm(f(xk)) - norm(f(xk) + jk'*p))
            if rhof < beta1 # linear approximation is poor fit so reduce the trust region
                deltak = min(0.25*deltak,0.5*norm(p))
                p .= step_selection(f,xk,Gk,jk,gk,pkn,l,u,deltak,theta)
            elseif rhof > beta2 # linear approximation is good fit so expand the trust region
                deltak = max(deltak,2*norm(p))
                p .= step_selection(f,xk,Gk,jk,gk,pkn,l,u,deltak,theta)
            end
            xn .= xk .+ p
            break
        end

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))

        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

    end
  
    results = SolverResults(:dogleg,x,xn,f(xn),lenx,lenf,iter)

    return results
 
end

function constrained_dogleg_solver_sparse(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100) where {T <: AbstractFloat, S <: Integer}

    # This is an implementation of the algorithm from Bellavia, Macconi, and Pieraccini (2012), "Constrained 
    # dogleg methods for nonlinear systems with simple bounds", Computational Optimization and Applications, 
    # 53, pp. 771--794 

    n = length(x)
    xk  = copy(x)
    xn  = similar(x)
    pkn = similar(x)
    p   = similar(x)
    df  = similar(x)

    jk = sparse(Array{T}(undef,n,n))
    Gk = similar(jk)
    gk = similar(x)

    lenx = zero(T)
    lenf = zero(T)

    # Replace infinities with largest possible Float64
    
    for i in eachindex(x)
        if l[i] == -Inf
            l[i] = -1/eps(T)
        elseif l[i] == Inf
            l[i] = 1/eps(T)
        end
        if u[i] == -Inf
            u[i] = -1/eps(T)
        elseif u[i] == Inf
            u[i] = 1/eps(T)
        end
    end
    
    deltak = 1.0 
    theta  = 0.99995 
    beta1  = 0.5#0.25 
    beta2  = 0.9 

    iter = 0
    while true

        jk .= sparse(ForwardDiff.jacobian(f,xk))
        df .= coleman_li(f,jk,xk,l,u)
        Gk .= sparse(Diagonal(df.^(-1/2)))
        gk .= -df.*jk'f(xk)
   
        alphak = max(theta,1.0-norm(f(xk)))
        pkn .= xk-jk\f(xk)
        box_projection!(pkn,l,u)
        pkn .= alphak*(pkn-xk)

        p .= step_selection(f,xk,Gk,jk,gk,pkn,l,u,deltak,theta)
    
        while true
            rhof = (norm(f(xk)) - norm(f(xk+p))) / (norm(f(xk)) - norm(f(xk) + jk'*p))
            if rhof < beta1 # linear approximation is poor fit so reduce the trust region
                deltak = min(0.25*deltak,0.5*norm(p))
                p .= step_selection(f,xk,Gk,jk,gk,pkn,l,u,deltak,theta)
            elseif rhof > beta2 # linear approximation is good fit so expand the trust region
                deltak = max(deltak,2*norm(p))
                p .= step_selection(f,xk,Gk,jk,gk,pkn,l,u,deltak,theta)
            end
            xn .= xk .+ p
            break
        end

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))

        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

    end
  
    results = SolverResults(:dogleg,x,xn,f(xn),lenx,lenf,iter)

    return results
 
end

function constrained_newton_krylov(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100,krylovdim::S=30) where {T <: AbstractFloat, S <: Integer}

    # This is an implementation of Algorithm 1 from Chen and Vuik (2016) "Globalization Technique for 
    # Projected Newton-Krylov Methods", International Journal  for Numerical Methods in Engineering, 
    # 110, pp. 661--674.

    n = length(x)
    xk = copy(x)
    xn = similar(x)
    xt = similar(x)

    jk = Array{T}(undef,n,n)

    lenx = zero(T)
    lenf = zero(T)

    g(x) = (1/2)*norm(f(x))^2

    etak = 1e-4
    beta = 0.9
    t = 1e-4
    sigma = 1e-4
    mmax = 50

    flag_ng = false
    iter = 0
    while true

        if flag_ng == false
            jk .= ForwardDiff.jacobian(f,xk)
            dk, status = gmres(jk,-f(xk),xk,etak,krylovdim)

            alpha = 1.0
            m = 1
            while m <= mmax
                xt .= xk+alpha*dk
                box_projection!(xt,l,u)
                if norm(f(xt)) <= (1.0 - t*alpha*(1-etak))*norm(f(xk))
                    xn .= xt
                    etak = (1.0 - alpha*(1.0 - etak))
                    flag_ng = false
                    break
                else
                    alpha = beta*alpha
                    m += 1
                    flag_ng = true
                end
            end
        else
            dk = -jk'f(xk)
            alpha = 1.0
            m = 1
            while m <= mmax
                xt .= xk+alpha*dk
                box_projection!(xt,l,u)
                if g(xt) <= g(xk) + sigma*dk'*(xt - xk)
                    xn .= xt
                    flag_ng = false
                    break
                else
                    alpha = beta*alpha
                    m += 1
                end
            end
        end

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))

        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

    end
  
    results = SolverResults(:nk,x,xn,f(xn),lenx,lenf,iter)

    return results
 
end

function constrained_newton_krylov_sparse(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100,krylovdim::S=30) where {T <: AbstractFloat, S <: Integer}

    # This is an implementation of Algorithm 1 from Chen and Vuik (2016) "Globalization Technique for 
    # Projected Newton-Krylov Methods", International Journal  for Numerical Methods in Engineering, 
    # 110, pp. 661--674.

    n = length(x)
    xk = copy(x)
    xn = similar(x)
    xt = similar(x)

    jk = sparse(Array{T}(undef,n,n))

    lenx = zero(T)
    lenf = zero(T)

    g(x) = (1/2)*norm(f(x))^2

    etak = 1e-4
    beta = 0.9
    t = 1e-4
    sigma = 1e-4
    mmax = 50

    flag_ng = false
    iter = 0
    while true

        if flag_ng == false
            jk .= sparse(ForwardDiff.jacobian(f,xk))
            dk, status = gmres(jk,-f(xk),xk,etak,krylovdim)

            alpha = 1.0
            m = 1
            while m <= mmax
                xt .= xk+alpha*dk
                box_projection!(xt,l,u)
                if norm(f(xt)) <= (1.0 - t*alpha*(1-etak))*norm(f(xk))
                    xn .= xt
                    etak = (1.0 - alpha*(1.0 - etak))
                    flag_ng = false
                    break
                else
                    alpha = beta*alpha
                    m += 1
                    flag_ng = true
                end
            end
        else
            dk = -jk'f(xk)
            alpha = 1.0
            m = 1
            while m <= mmax
                xt .= xk+alpha*dk
                box_projection!(xt,l,u)
                if g(xt) <= g(xk) + sigma*dk'*(xt - xk)
                    xn .= xt
                    flag_ng = false
                    break
                else
                    alpha = beta*alpha
                    m += 1
                end
            end
        end

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))

        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

    end
  
    results = SolverResults(:nk,x,xn,f(xn),lenx,lenf,iter)

    return results
 
end

function constrained_newton_krylov_fs(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100,krylovdim::S=30) where {T <: AbstractFloat, S <: Integer}

    # This is an implementation of the approach in Frontini and Sormani (2004) "Third-order methods from
    # quadrature formulae for solving systems of nonlinear equations", Applied Mathematics and Computation, 
    # 149, pp. 771--782.

    # Modified to use Krylov methods, a globalization step, and to allow for box-constraints by Richard Dennis.

    n = length(x)
    xk = copy(x)
    xn = similar(x)
    xt = similar(x)

    lenx = zero(T)
    lenf = zero(T)

    g(x) = (1/2)*norm(f(x))^2

    jk = Array{T}(undef,n,n)
     
    etak = 1e-4
    beta = 0.9
    t = 1e-4
    sigma = 1e-4
    mmax = 50

    flag_ng = false
    iter = 0
    while true

        if flag_ng == false

            jk .= ForwardDiff.jacobian(f,xk)
            kk, status = gmres(jk,-(1/2)*f(xk),xk,etak,krylovdim)
            jk .= ForwardDiff.jacobian(f,xk-kk) # Here xk-kk is not guaranteed to be inside the box
            dk, status = gmres(jk,-f(xk),xk,etak,krylovdim)

            alpha = 1.0
            m = 1
            while m <= mmax
                xt .= xk+alpha*dk
                box_projection!(xt,l,u)
                if norm(f(xt)) <= (1.0 - t*alpha*(1-etak))*norm(f(xk))
                    xn .= xt
                    etak = (1.0 - alpha*(1.0 - etak))
                    flag_ng = false
                    break
                else
                    alpha = beta*alpha
                    m += 1
                    flag_ng = true
                end
            end
        else
            dk = -jk'f(xk)
            alpha = 1.0
            m = 1
            while m <= mmax
                xt .= xk+alpha*dk
                box_projection!(xt,l,u)
                if g(xt) <= g(xk) + sigma*dk'*(xt - xk)
                    xn .= xt
                    flag_ng = false
                    break
                else
                    alpha = beta*alpha
                    m += 1
                end
            end
        end

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))

        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

    end
  
    results = SolverResults(:nk_fs,x,xn,f(xn),lenx,lenf,iter)

    return results
 
end

function constrained_newton_krylov_fs_sparse(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100,krylovdim::S=30) where {T <: AbstractFloat, S <: Integer}

    # This is an implementation of the approach in Frontini and Sormani (2004) "Third-order methods from
    # quadrature formulae for solving systems of nonlinear equations", Applied Mathematics and Computation, 
    # 149, pp. 771--782.

    # Modified to use Krylov methods, a globalization step, and to allow for box-constraints by Richard Dennis.

    n = length(x)
    xk = copy(x)
    xn = similar(x)
    xt = similar(x)

    lenx = zero(T)
    lenf = zero(T)

    g(x) = (1/2)*norm(f(x))^2

    jk = sparse(Array{T}(undef,n,n))
     
    etak = 1e-4
    beta = 0.9
    t = 1e-4
    sigma = 1e-4
    mmax = 50

    flag_ng = false
    iter = 0
    while true

        if flag_ng == false

            jk .= sparse(ForwardDiff.jacobian(f,xk))
            kk, status = gmres(jk,-(1/2)*f(xk),xk,etak,krylovdim)
            jk .= sparse(ForwardDiff.jacobian(f,xk-kk)) # Here xk-kk is not guaranteed to be inside the box
            dk, status = gmres(jk,-f(xk),xk,etak,krylovdim)

            alpha = 1.0
            m = 1
            while m <= mmax
                xt .= xk+alpha*dk
                box_projection!(xt,l,u)
                if norm(f(xt)) <= (1.0 - t*alpha*(1-etak))*norm(f(xk))
                    xn .= xt
                    etak = (1.0 - alpha*(1.0 - etak))
                    flag_ng = false
                    break
                else
                    alpha = beta*alpha
                    m += 1
                    flag_ng = true
                end
            end
        else
            dk = -jk'f(xk)
            alpha = 1.0
            m = 1
            while m <= mmax
                xt .= xk+alpha*dk
                box_projection!(xt,l,u)
                if g(xt) <= g(xk) + sigma*dk'*(xt - xk)
                    xn .= xt
                    flag_ng = false
                    break
                else
                    alpha = beta*alpha
                    m += 1
                end
            end
        end

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))

        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

    end
  
    results = SolverResults(:nk_fs,x,xn,f(xn),lenx,lenf,iter)

    return results
 
end

function constrained_jacobian_free_newton_krylov(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100,krylovdim::S=30) where {T <: AbstractFloat, S <: Integer}

    # This function is my implementation of a constrained Jacobian-free Newton-Krylov solver

    xk = copy(x)
    xn = similar(x)
    dk = similar(x)

    lenx = zero(T)
    lenf = zero(T)

    g(x) = (1/2)*norm(f(x))^2

    etak = 1e-4
    beta = 0.9
    t = 1e-4
    sigma = 1e-4
    mmax = 50

    flag_ng = false
    iter = 0
    while true

        if flag_ng == false
            dk, status = jacobian_free_gmres(f,xk,etak,krylovdim) # Using inexact Arnoldi (could use restarted)

            if status == false
                error("Restarted GMRES did not converge")
            end

            alpha = 1.0
            m = 1
            while m <= mmax
                if norm(f(box_projection(xk+alpha*dk,l,u))) <= (1.0 - t*alpha*(1-etak))*norm(f(xk))
                    xn = box_projection(xk+alpha*dk,l,u)
                    etak = (1.0 - alpha*(1.0 - etak))
                    flag_ng = false
                    break
                else
                    alpha = beta*alpha
                    m += 1
                    flag_ng = true
                end
            end
        else
            dk = - ForwardDiff.gradient(g,xk)
            alpha = 1.0
            m = 1
            while m <= mmax
                if g(box_projection(xk+alpha*dk,l,u)) <= g(xk) + sigma*dk'*(box_projection(xk+alpha*dk,l,u) .- xk)
                    xn = box_projection(xk+alpha*dk,l,u)
                    flag_ng = false
                    break
                else
                    alpha = beta*alpha
                    m += 1
                end
            end
        end

        lenx = maximum(abs,xn-xk)
        lenf = maximum(abs,f(xn))

        xk .= xn

        iter += 1

        if iter >= maxiters || (lenx <= xtol || lenf <= ftol)
            break
        end

    end
  
    results = SolverResults(:jfnk,x,xn,f(xn),lenx,lenf,iter)

    return results
 
end

function nlboxsolve(f::Function,x::Array{T,1},l::Array{T,1},u::Array{T,1};xtol::T=1e-8,ftol::T=1e-8,maxiters::S=100,method::Symbol = :lm_ar,krylovdim::S=30,sparsejac::Symbol=:no) where {T <: AbstractFloat, S <: Integer}

    if method == :newton && sparsejac == :no
        return constrained_newton(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters)
    elseif method == :lm && sparsejac == :no
        return constrained_levenberg_marquardt(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters)
    elseif method == :lm_kyf && sparsejac == :no
        return constrained_levenberg_marquardt_kyf(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters)
    elseif method == :lm_fan && sparsejac == :no
        return constrained_levenberg_marquardt_fan(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters)
    elseif method == :lm_ar && sparsejac == :no
        return constrained_levenberg_marquardt_ar(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters)
    elseif method == :dogleg && sparsejac == :no
        return constrained_dogleg_solver(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters)
    elseif method == :nk && sparsejac == :no
        return constrained_newton_krylov(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters,krylovdim=krylovdim)
    elseif method == :nk_fs && sparsejac == :no
        return constrained_newton_krylov_fs(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters,krylovdim=krylovdim)
    elseif method == :newton && sparsejac == :yes
        return constrained_newton_sparse(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters)
    elseif method == :lm && sparsejac == :yes
        return constrained_levenberg_marquardt_sparse(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters)
    elseif method == :lm_kyf && sparsejac == :yes
        return constrained_levenberg_marquardt_kyf_sparse(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters)
    elseif method == :lm_fan && sparsejac == :yes
        return constrained_levenberg_marquardt_fan_sparse(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters)
    elseif method == :lm_ar && sparsejac == :yes
        return constrained_levenberg_marquardt_ar_sparse(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters)
    elseif method == :dogleg && sparsejac == :yes
        return constrained_dogleg_solver_sparse(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters)
    elseif method == :nk && sparsejac == :yes
        return constrained_newton_krylov_sparse(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters,krylovdim=krylovdim)
    elseif method == :nk_fs && sparsejac == :yes
        return constrained_newton_krylov_fs_sparse(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters,krylovdim=krylovdim)
    elseif method == :jfnk
        return constrained_jacobian_free_newton_krylov(f,x,l,u,xtol=xtol,ftol=ftol,maxiters=maxiters,krylovdim=krylovdim)
    end

end
